-- Puffin: A Tojaso FowlPlay Project
-- Data management functions based on a three-level hierarchy: database, record, attribute
-- Attributes are name-value pairs (values can be tables with memory managed by caller)
-- Records can be redirected to access attributes from a caller-provided table
-- Optionally, a validate function can be called before accessing a data field to make sure it is valid

local Puffin_Data = {}
local databases = {}
local empty = {}

-- Allocate a new database, initialize records and set persistent flag
function Puffin_Data.AllocateDatabase(name, persistent)
	local dt = databases[name]
	if not dt then dt = Puffin.AllocateTable(); dt._records = Puffin.AllocateTable(); dt._redirect = Puffin.AllocateTable(); databases[name] = dt end
	dt._persistent = persistent
end

-- Release and recycle a database, release all tables used to store records and attributes
function Puffin_Data.ReleaseDatabase(name)
	local dt = databases[name]
	if dt then
		for k, v in pairs(dt._records) do dt._records[k] = Puffin.ReleaseTable(v) end
		Puffin.ReleaseTable(dt._records); Puffin.ReleaseTable(dt._redirect); Puffin.ReleaseTable(dt)
	end
	databases[name] = nil
end

-- Release all current records in a database
function Puffin_Data.ReleaseRecords(name)
	local dt = databases[name]
	if dt then
		for k, v in pairs(dt._records) do dt._records[k] = Puffin.ReleaseTable(v) end
		table.wipe(dt._redirect)
	end
end

-- Return values equivalent to pairs() that can be used to iterate over all non-redirected records in the database
function Puffin_Data.GetRecords(name)
	local dt = databases[name]
	if dt then return pairs(dt._records) end
	return pairs(empty)
end

-- Set attributes in a database record, optionally wiping current contents before updating
function Puffin_Data.SetRecord(name, record, atable, wipe)
	local dt = databases[name]
	if dt and record and atable then
		local dr = dt._records[record]
		if not dr then dr = Puffin.AllocateTable(); dt._records[record] = dr end
		if wipe then table.wipe(dr) end -- first delete current contents
		for key, value in pairs(atable) do dr[key] = value end
		dt._redirect[record] = nil -- removing any redirect set for this record
	end
end

-- Set a redirection table for a record, attributes will be read from this table instead of the record's internal table
-- Memory management for the redirect table is the responsibility of the caller
function Puffin_Data.SetRedirect(name, record, atable)
	local dt = databases[name]
	if dt and record and atable then
		local dr = dt._records[record]
		if dr then dt._records[record] = Puffin.ReleaseTable(dr) end --release old record attribute table
		dt._redirect[record] = atable
	end
end

-- Set a validation function for a database to be called before returning a record or attribute
function Puffin_Data.SetValidate(name, validate)
	local dt = databases[name]
	if dt and type(validate) == "function" then dt._validate = validate end
end

-- Get the attributes in a database record
function Puffin_Data.GetRecord(name, record)
	local dt = databases[name]
	if dt and record then
		if dt._validate then dt._validate(name, record) end -- optional validate function called before returning the record
		local dr = dt._records[record]; if not dr then dr = dt._redirect[record] end
		return dr
	end
	return nil
end

-- Release a database record, return true only if there was a record to be released
function Puffin_Data.ReleaseRecord(name, record)
	local dt = databases[name]
	if dt and record then
		local dr = dt._records[record]; if dr then dt._records[record] = Puffin.ReleaseTable(dr); return true end
		dr = dt._redirect[record]; if dr then dt._redirect[record] = nil; return true end
	end
	return nil
end

-- Set an attribute's value in a database record, following redirect if any and creating a record table if necessary
function Puffin_Data.SetAttribute(name, record, attribute, value)
	local dt = databases[name]
	if dt and record then
		local dr = dt._records[record]
		if not dr then dr = dt._redirect[record] end -- use redirect table if set
		if not dr then dr = Puffin.AllocateTable(); dt._records[record] = dr end
		dr[attribute] = value
	end
end

-- Get the value of an attribute in a database record, following redirect if any
function Puffin_Data.GetAttribute(name, record, attribute)
	local dt = databases[name]
	if dt and record then
		if dt._validate then dt._validate(name, record, attribute) end -- optional validate function called before returning the attribute
		local dr = dt._records[record]; if not dr then dr = dt._redirect[record] end -- use redirect table if set
		if dr then return dr[attribute] end
	end
	return nil
end

Puffin.RegisterComponent("Data", Puffin_Data)