-- Puffin: A Tojaso FowlPlay Project
-- Model component for monitoring spell casting for player and target, plus mirror timers

local Model_Casting = {}
local DB = nil
local unit_changed = {}
local unit_tracking = {}
local spellTicks, localSpells, _
local mirrorIcons = {}
local sentTime = nil
local latency = 0

local function CastingEvent(e, unit) if unit == "vehicle" then unit = "player" end if unit_tracking[unit] then unit_changed[unit] = true end end
local function CastingUnitEvent(e, unit, track) unit_tracking[unit] = track; unit_changed[unit] = true end
local function CastingSentEvent(e, unit) if unit == "player" or unit == "vehicle" then sentTime = GetTime() end end
local function CastingStartEvent(e, unit) if unit == "player" or unit == "vehicle" and sentTime then latency = GetTime() - sentTime end CastingEvent(e, unit) end
local function MirrorEvent(e) if unit_tracking.mirror then unit_changed.mirror = true end end
local function CastingTargetEvent(e) CastingEvent(e, "target") end
local function CastingFocusEvent(e) CastingEvent(e, "focus") end
local function CastingPetEvent(e) CastingEvent(e, "pet") end

-- Initialize by creating a database for cooldowns (assumes records will not be redirected)
function Model_Casting.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, "Puffin.Model_Casting: invalid reference to data management component")
	DB.AllocateDatabase("casting", false)
	spellTicks = DB.GetRecord("spells", "ticks")
	assert(spellTicks, "Puffin.Model_Casting: missing spell ticks")
	localSpells = DB.GetRecord("spells", "local")
	assert(localSpells, "Puffin.Model_Casting: missing local spells")
	
	_, _, mirrorIcons.BREATH = GetSpellInfo(5697)
	_, _, mirrorIcons.DEATH = GetSpellInfo(98391)
	_, _, mirrorIcons.EXHAUSTION = GetSpellInfo(57723)
	_, _, mirrorIcons.FEIGNDEATH = GetSpellInfo(5384)

	Puffin.RegisterCallback("UNIT_SPELLCAST_START", CastingStartEvent) -- register for events that affect casting
	Puffin.RegisterCallback("UNIT_SPELLCAST_STOP", CastingEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_SENT", CastingSentEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_DELAYED", CastingEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_FAILED", CastingEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_INTERRUPTED", CastingEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_INTERRUPTIBLE", CastingEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_NOT_INTERRUPTIBLE", CastingEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_CHANNEL_START", CastingEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_CHANNEL_STOP", CastingEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_CHANNEL_UPDATE", CastingEvent)
	Puffin.RegisterCallback("MIRROR_TIMER_START", MirrorEvent)
	Puffin.RegisterCallback("MIRROR_TIMER_STOP", MirrorEvent)
	Puffin.RegisterCallback("MIRROR_TIMER_PAUSE", MirrorEvent)
	Puffin.RegisterCallback("CastingUnitTracking", CastingUnitEvent)
	Puffin.RegisterCallback("PLAYER_TARGET_CHANGED", CastingTargetEvent)
	Puffin.RegisterCallback("PLAYER_FOCUS_CHANGED", CastingFocusEvent)
	Puffin.RegisterCallback("UNIT_PET", CastingPetEvent)
end

-- On update, manage database of active cooldowns in response to cooldown-related events
function Model_Casting.OnUpdate()
	for unit, changed in pairs(unit_changed) do
		if changed and unit_tracking[unit] and unit ~= "mirror" then
			unit_changed[unit] = false
			local checkUnit = unit
			if UnitHasVehicleUI("player") then
				if unit == "player" then checkUnit = "pet" elseif unit == "pet" then checkUnit = "player" end
			end
			local channel, fillRate = false, 1
			local name, _, label, icon, startTime, endTime, tradeSkill, _, interrupt = UnitCastingInfo(checkUnit)
			if not name then
				name, _, label, icon, startTime, endTime, tradeSkill, interrupt = UnitChannelInfo(checkUnit)
				channel = true; fillRate = -1; label = name
			end
			if name then
				DB.SetAttribute("casting", unit, "label", label)
				DB.SetAttribute("casting", unit, "icon", icon)
				DB.SetAttribute("casting", unit, "expire", endTime / 1000)
				DB.SetAttribute("casting", unit, "duration", (endTime - startTime) / 1000)
				DB.SetAttribute("casting", unit, "tradeSkill", tradeSkill)
				DB.SetAttribute("casting", unit, "interrupt", interrupt)
				DB.SetAttribute("casting", unit, "channel", channel)
				if not channel and unit == "player" then DB.SetAttribute("casting", unit, "latency", latency) end
				local spell, ticks = localSpells[name], 0
				local t = spellTicks[spell]; if t then ticks = t.ticks or 0 end
				DB.SetAttribute("casting", unit, "ticks", ticks)
				-- print("casting", unit, name, label, icon, (endTime - startTime) / 1000, endTime / 1000, tradeSkill, interrupt, channel, ticks, latency)
			else
				DB.ReleaseRecord("casting", unit)
			end
			Puffin.Callback("Casting", unit)
		end
	end

	if unit_changed.mirror then
		local found = false; unit_changed.mirror = false
		if unit_tracking.mirror then
			for i = 1, MIRRORTIMER_NUMTIMERS do
				local timer, timeLeft, duration, fillRate, paused, label = GetMirrorTimerInfo(i)
				if timeLeft and timeLeft > 0 then
					local icon = mirrorIcons[timer]
					DB.SetAttribute("casting", "mirror", "label", label)
					DB.SetAttribute("casting", "mirror", "icon", icon)
					DB.SetAttribute("casting", "mirror", "expire", (timeLeft / 1000) + GetTime())
					DB.SetAttribute("casting", "mirror", "duration", duration / 1000)
					DB.SetAttribute("casting", "mirror", "fillRate", fillRate)
					if fillRate > 1 then DB.SetAttribute("casting", "mirror", "timeLeft", timeLeft / 1000) end
					DB.SetAttribute("casting", "mirror", "paused", paused)
					DB.SetAttribute("casting", "mirror", "timer", timer)
					DB.SetAttribute("casting", "mirror", "channel", true)
					-- print("mirror", i, timer, label, icon, duration / 1000, (timeLeft / 1000) + GetTime(), fillRate, paused)
					found = true
					break -- only one active mirror timer at a time
				end
			end
		end
		if not found then DB.ReleaseRecord("casting", "mirror") end
		Puffin.Callback("Casting", "mirror")
	end
end

-- On refresh, update progress for any active mirror castbar
function Model_Casting.OnRefresh()
	local fillRate = DB.GetAttribute("casting", "mirror", "fillRate")
	if fillRate and fillRate > 1 then
		local timer = DB.GetAttribute("casting", "mirror", "timer")
		local timeLeft = GetMirrorTimerProgress(timer)
		if timeLeft and timeLeft > 0 then
			DB.SetAttribute("casting", "mirror", "timeLeft", timeLeft / 1000)
			Puffin.Callback("Casting", "mirror")
		end
	end
end

Puffin.RegisterComponent("Model_Casting", Model_Casting, "MODEL", 5)
